#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains a set of utility functions which are used by the
other modules in the `linalg` package. The functions defined herein
operate on matrices, or are at the very least related to linear algebra
computations.
"""

import functools
from functools import partial
from sympy import Matrix, gcd, lcm
from abelian.utils import mod


def columns_as_list(A):
    """
    Returns the columns of A as a list of lists.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix.

    Returns
    -------
    list_of_cols : list
        A list of lists, where each sub_list is a column,
        e.g. structure [[col1], [col2], ...].

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[1, 2],
    ...             [3, 4]])
    >>> list_of_cols = columns_as_list(A)
    >>> list_of_cols
    [[1, 3], [2, 4]]
    """
    m, n = A.shape
    list_of_cols = [[A[i, j] for i in range(0, m)] for j in range(0, n)]
    return list_of_cols


def nonzero_columns(H):
    """
    Counts the number of columns in H not identically zero.

    Parameters
    ----------
    H : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix.

    Returns
    -------
    nonzero_cols : int
        The number of columns of A not indentically zero.

    Examples
    ---------
    >>> from sympy import Matrix, diag
    >>> A = Matrix([[0, 2],
    ...             [0, 4]])
    >>> nonzero_columns(A)
    1
    >>> nonzero_columns(Matrix.eye(5))
    5
    >>> nonzero_columns(diag(0,1,0,3,5,0))
    3
    """
    m, n = H.shape
    nonzero_cols = 0
    # Loop over the columns
    for j in range(0, n):

        # Loop over the rows
        for i in range(0, m):
            if H[i, j] != 0:
                nonzero_cols += 1
                break

    return nonzero_cols


def diagonal_rank(S):
    """
    Count the number of non-zero diagonals in S,
    where S is in Smith normal form.

    Parameters
    ----------
    S : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix in Smith normal form.

    Returns
    -------
    num_nonzeros : int
        The number of non-zeros on the diagonal.

    Examples
    ---------
    >>> from sympy import diag
    >>> diagonal_rank(diag(1,2,0,0,0,0))
    2
    >>> diagonal_rank(diag(1,2,4,8,0,0))
    4
    """
    return len(nonzero_diag_as_list(S))


def nonzero_diag_as_list(S):
    """
    Return a list of the non-zero diagonals entries of S.

    Parameters
    ----------
    S : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix, typically in Smith normal form.

    Returns
    -------
    nonzero_diags : list
        A list of the non-zero diagonal entries of S.

    Examples
    ---------
    >>> from sympy import diag
    >>> nonzero_diag_as_list(diag(1,2,0,0,0,0))
    [1, 2]
    >>> nonzero_diag_as_list(diag(1,2,4,8,0,0))
    [1, 2, 4, 8]
    """
    m, n = S.shape
    nonzero_diags = [S[i, i] for i in range(0, min(m, n)) if S[i, i] != 0]
    return nonzero_diags


def remove_zero_columns(M):
    """
    Return a copy of M where the columns that are identically zero are deleted.

    Parameters
    ----------
    M : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix with zero or more columns which are identically zero.

    Returns
    -------
    M : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A copy of the input matrix with all zero columns removed.

    Examples
    ---------
    >>> from sympy import Matrix, diag
    >>> A = Matrix([[0, 1],
    ...             [0, 2]])
    >>> remove_zero_columns(A) == Matrix([1, 2])
    True
    >>> A = diag(0,1,2)
    >>> A_del = Matrix([[0, 0],
    ...                 [1, 0],
    ...                 [0, 2]])
    >>> remove_zero_columns(A) == A_del
    True
    """

    A = M.copy()
    m, n = A.shape

    # STEP 1: Find the columns to remove
    cols_to_remove = []

    # Iterate over columns
    for j in range(0, n):
        all_zero = True
        for i in range(0, m):
            if A[i, j] != 0:
                all_zero = False
                break

        # If the column has all zeros, we delete it later
        if all_zero:
            cols_to_remove.append(j)

    # STEP 2: Remove the columns and return
    return remove_cols(A, cols_to_remove)


def remove_cols(A, cols_to_remove):
    """
    Return a copy of A where the columns with indices in `cols_to_remove`
    are removed.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix.

    cols_to_remove : list
        A list of column indices to remove from `A`.

    Returns
    -------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A copy of the input matrix with removed columns.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[5, 6, 7, 8]])
    >>> B = remove_cols(A, [0, 2])
    >>> B == Matrix([[6, 8]])
    True
    """
    m, n = A.shape
    if any([index >= n for index in cols_to_remove]):
        raise ValueError('Index to remove not in matrix.')

    cols_to_remove.sort()

    new_A = A.copy()
    deleted = 0
    for j in cols_to_remove:
        new_A.col_del(j - deleted)
        deleted += 1

    return new_A


def remove_rows(A, rows_to_remove):
    """
    Return a copy of A where the rows with indices in `rows_to_remove`
    are removed.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix.

    rows_to_remove : list
        A list of row indices to remove from `A`.

    Returns
    -------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A copy of the input matrix with removed rows.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[5, 6, 7, 8]]).T
    >>> B = remove_rows(A, [0, 2])
    >>> B == Matrix([[6, 8]]).T
    True
    """
    return remove_cols(A.T, rows_to_remove).T

def vector_mod_vector(vector, mod_vector):
    """
    Return `vector` % `mod_vector`, a vectorized mod operation.

    Parameters
    ----------
    vector : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy column vector, i.e. a sympy matrix of dimension m x 1.
    mod_vector : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy column vector, i.e. a sympy matrix of dimension m x 1.

    Returns
    -------
    modded_vector : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The result of the mod operation on every entry.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> element = Matrix([5, 7, 9])
    >>> mod_vect = Matrix([3, 3, 5])
    >>> modded = vector_mod_vector(element, mod_vect)
    >>> modded == Matrix([2, 1, 4])
    True
    """

    if len(vector) != len(mod_vector):
        raise ValueError('Vectors must be of same length.')

    entries = (mod(i, j) for (i,j) in zip(vector, mod_vector))
    modded_vector = Matrix(list(entries))
    return modded_vector


def matrix_mod_vector(A, mod_col):
    """
    Returns a copy of `A` with every column modded by `mod_col`.

    Parameters
    ----------
    vector : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix of size m x n.
    mod_col : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy column vector, i.e. a sympy matrix of dimension m x 1.

    Returns
    -------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A copy of the input with each column modded.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[5, 6],
    ...             [8, 5],
    ...             [3, 5]])
    >>> mod_col = Matrix([4, 6, 3])
    >>> A_modded = matrix_mod_vector(A, mod_col)
    >>> A_modded == Matrix([[1, 2],
    ...                     [2, 5], [0, 2]])
    True
    """
    m, n = A.shape
    if m != len(mod_col):
        raise ValueError('Dimension mismatch.')

    new_A = A.copy()

    # If it's a zero matrix (point),
    # return immediately.
    if m * n == 0:
        return new_A

    for j in range(0, n):
        column = A[:, j]
        new_A[:, j] = vector_mod_vector(column, mod_col)

    return new_A


def order_of_vector(v, mod_vector):
    """
    Returns the order of the element `v` in a FGA like `mod_vector`.

    Parameters
    ----------
    v : :py:class:`~sympy.matrices.dense.MutableDenseMatrix` or a list
        An iterable object with integers. This is the group element.
    mod_vector : :py:class:`~sympy.matrices.dense.MutableDenseMatrix` or a list
        An iterable object with integers. This is the orders of the group.

    Returns
    -------
    order : int
        The order of `v` in `mod_vector`.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> order_of_vector([1,2,3], [2,4,6]) # Order of 2
    2
    >>> order_of_vector([1,2,3], [0, 0, 0]) # Order of 0 (infinite order)
    0
    >>> order_of_vector([1,2,3], [7, 5, 2]) # lcm(7, 10, 2) is 70
    70
    >>> order_of_vector([0,0,0], [0,0,0]) # Identity element
    1
    >>> order_of_vector([0,2, 3], [0,0,0]) # Non-trivial element
    0
    >>> order_of_vector([1, 0, 1], [5, 0, 0])
    0

    """

    # Arguments must have the same length
    if len(v) != len(mod_vector):
        raise ValueError('The arguments must have the same length.')

    def div(top, bottom):
        # Division, but if we divide by zero we define it as unity
        try:
            return top // bottom
        except:
            return 1

    # Create the generate and iterate through it
    gen = zip(v, mod_vector)
    gcd_list = (div(order, gcd(element, order)) for(element, order) in gen)

    # Reduce over the iterator with the least common multiple
    return functools.reduce(lcm, gcd_list)


def mat_times_diag(A, diagonal):
    """
    Multiply a dense matrix and a diagonal.

    Multiplies `A` with a column vector `diagonal`, which is interpreted as
    the diagonal of a matrix. This algorithm exploids the diagonal structure
    to reduce the number of computations.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A dense matrix.
    diag : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The diagonal of a matrix, represented as a sympy column vector.

    Returns
    -------
    product : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The product `A` times `diag`.

    Examples
    ---------
    >>> from sympy import Matrix, diag
    >>> A = Matrix([[1, 2],
    ...             [3, 4]])
    >>> diagonal = Matrix([2, 3])
    >>> mat_times_diag(A, diagonal) == A * diag(2, 3)
    True
    """
    m, n = A.shape
    new_A = A.copy()
    for col in range(0, n):
        new_A[:, col] *= diagonal[col]
    return new_A


def diag_times_mat(diagonal, A):
    """
    Multiply a diagonal and a dense matrix.

    Multiplies a column vector `diagonal` with `A`, in that order.
    This algorithm exploids the diagonal structure
    to reduce the number of computations.

    Parameters
    ----------
    diag : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The diagonal of a matrix, represented as a sympy column vector.
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A dense matrix.

    Returns
    -------
    product : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The product `diag` times `A`.

    Examples
    ---------
    >>> from sympy import Matrix, diag
    >>> A = Matrix([[1, 2],
    ...             [3, 4]])
    >>> diagonal = Matrix([2, 3])
    >>> diag_times_mat(diagonal, A) == diag(2, 3) * A
    True
    """

    return mat_times_diag(A.T, diagonal).T


def reciprocal_entrywise(A):
    """
    Returns the entrywise reciprocal of a matrix or vector.

    Will skip zero entries.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix, or vector ( m x 1 matrix).

    Returns
    -------
    reciprocal : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The entrywise reciprocal of `A`.

    Examples
    ---------
    >>> from sympy import Matrix, diag
    >>> D = diag(1, 2, 3)
    >>> D_inv = reciprocal_entrywise(D)
    >>> D * D_inv == Matrix.eye(3)
    True
    >>> A = Matrix([[1, 5], [4, 1]])
    >>> A_recip = reciprocal_entrywise(A)
    >>> A_recip == Matrix([[1, 1/5], [1/4, 1]])
    True
    """
    m, n = A.shape
    def recip(x):
        if x == 0:
            return x
        return 1/x

    return Matrix(m, n, [recip(e) for e in A])


def norm(vector, p = 2):
    """
    The p-norm of an iterable.

    Parameters
    ----------
    vector : :py:class:`~sympy.matrices.dense.MutableDenseMatrix` or list
        The iterable to compute the norm over.
    p : float
        The p-value in the p-norm. Should be between 1 and infinity (None).

    Returns
    -------
    norm : float
        The computed norm.

    Examples
    --------
    >>> vector = [1, 2, 3]
    >>> norm(vector, 1)
    6.0
    >>> norm(tuple(vector), None)
    3.0
    >>> norm(iter(vector), None)
    3.0
    >>> norm(vector, None)
    3.0
    >>> norm(vector, 2)
    3.7416573867739413
    >>> from sympy import Matrix
    >>> vector = Matrix(vector)
    >>> norm(vector, 1)
    6.0
    >>> norm(vector, None)
    3.0
    >>> norm(vector, 2)
    3.7416573867739413

    """

    # If no p is given, assume the infinity norm and return the max value
    if p == None:
        return float(max(abs(j) for j in list(vector)))

    # P should not be smaller than 1.
    if p < 1:
        raise ValueError('p must be >= 1 for the p-norm.')

    # Return the p-norm
    norm = sum(abs(i)**p for i in vector)**(1/p)
    return float(norm)


def difference(iterable1, iterable2, p = None):
    """
    Compute the difference with a p-norm.

    Parameters
    ----------
    iterable1 : :py:class:`~sympy.matrices.dense.MutableDenseMatrix` or list
        The iterable to compute the norm over.
    iterable2 : :py:class:`~sympy.matrices.dense.MutableDenseMatrix` or list
        The iterable to compute the norm over.
    p : float
        The p-value in the p-norm. Should be between 1 and infinity (None).

    Returns
    -------
    norm : float
        The computed norm of the difference.

    Examples
    --------
    >>> 2 + 2
    4

    """
    vector1 = Matrix(iterable1)
    vector2 = Matrix(iterable2)
    diff = vector1 - vector2
    return norm(diff, p = p)


# Use partial functions to define common norms
euc_norm = partial(norm, p = 2)
max_norm = partial(norm, p = None)
one_norm = partial(norm, p = 1)


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose = False)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains equation solvers. All the inputs and outputs are of type
:py:class:`~sympy.matrices.dense.MutableDenseMatrix`.
"""

from sympy import gcdex, Matrix, diag
from abelian.linalg.free_to_free import free_kernel
from abelian.linalg.utils import remove_zero_columns, remove_cols


def solve(A, b, p = None):
    """
    Solve eqn Ax = b mod p over Z.

    The data (A, b, p) must be integer. The equation Ax = b mod p is solved,
    if a solution exists. If A is an epimorphism but not a monomorphism (i.e.
    overdetermined), one of the possible solutions is returned. If A is a
    monomorphism but not an epimorphism (i.e. underdetermined), a solution
    will be returned if one exists. If there is no solution, None is returned.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy integer matrix of size m x n.

    b : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy column matrix of size m x 1.

    p : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy column matrix of size m x 1. This column matrix represents
        the orders of the target group of A. If None, p will be set to the
        zero vector, i.e. infinite order in all components.

    Returns
    -------
    x : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A solution to A*x = b mod p, where x is of size n x 1.
        If no solution is found, None is returned.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> from abelian.linalg.utils import vector_mod_vector
    >>> A = Matrix([[5, 0, 3],
    ...             [0, 3, 4]])
    >>> x = Matrix([2, -1, 2])
    >>> p = Matrix([9, 9])
    >>> b = vector_mod_vector(A*x, p)
    >>> x_sol = solve(A, b, p)
    >>> vector_mod_vector(A*x_sol, p) == b
    True
    """

    # If no orders are supplied by the user, set the orders to zero,
    # i.e. infinite order or free-to-free.
    if p is None:
        m, n = b.shape
        p = Matrix(m, n, lambda i, j: 0)

    # Verify that the dimensions make sense
    (A_rows, A_cols) = A.shape
    (b_rows, b_cols) = b.shape
    (p_rows, p_cols) = p.shape
    if not (A_rows == b_rows == p_rows):
        raise ValueError('Dimension mismatch.')

    # Find the kernel of the projection onto the space Z_`p`
    ker_pi = remove_zero_columns(diag(*p))

    # Stack A | ker(pi) | b
    joined_A_D_b = A.row_join(ker_pi).row_join(b)

    # Compute ker( A | ker(pi) | b)
    kernel = free_kernel(joined_A_D_b)

    # The solution must be a linear combination of the columns of
    # ker( A | ker(pi) | b) such that the resulting vector has a -1
    # in the bottom entry.

    # Remove all columns with zero in the bottom entry
    m, n = kernel.shape
    col_indices = [j for j in range(n) if kernel[-1, j] == 0]
    kernel = remove_cols(kernel, col_indices)

    # Return None if the kernel is empty
    m, n = kernel.shape
    if n == 0:
        return None

    # Iteratively 'collapse' the columns using the extended
    # euclidean algorithm till the result is 1.
    m, n = kernel.shape
    while n > 1:
        # Compute the new column from the first two current ones
        f, g = kernel[-1, 0], kernel[-1, 1]
        (s, t, h) = gcdex(f, g)  # s*f + t*g = h.
        new_col = s * kernel[:, 0] + t * kernel[:, 1]

        # If there are only two columns, we have found the kernel
        if n == 2:
            kernel = new_col
            break

        # Delete current columns and insert the new one
        kernel = remove_cols(kernel, [0, 1])
        kernel = new_col.row_join(kernel)

        # Calculate new n value for the while-loop
        (m, n) = kernel.shape

    # Find shape of input, since shape of output depends on it
    (m, n) = A.shape

    # Make sure that the bottom row is -1 or 1.
    # It will always be 1 if the above while loop initiated,
    # but if it never initiated then value could be -1
    if kernel[-1, 0] not in [1, -1]:
        return None

    # The solution tot he problem is contained the first n rows of the
    # kernel, which is a column vector. Multiply by -1 if needed to
    # make sure the bottom entry is -1
    if kernel[-1, 0] == 1:
        return -kernel[:n, 0]
    else:
        return kernel[:n, 0]


def solve_epi(A, B, p = None):
    """
    Solve the equation X * mod p * A = B, where A is an epimorphism.

    The algorithm will produce a solution if (mod p * A) has a one
    sided inverse such that A_inv * A = I, i.e. A is an epimorphism.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy integer matrix of size m x n.

    B : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy column matrix of size k x n.

    p : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy column matrix of size m x 1. This column matrix represents
        the orders of the target group of A. If None, p will be set to the
        zero vector, i.e. infinite order.

    Returns
    -------
    x : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A solution to X * mod p * A = B.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> from abelian.linalg.utils import vector_mod_vector
    >>> A = Matrix([[5, 0, 3],
    ...             [0, 3, 4]])
    >>> X = Matrix([[1, 1],
    ...             [0, 1]])
    >>> B = X * A
    >>> X_sol = solve_epi(A, B)
    >>> X_sol * A == B
    True
    """
    # If no p (orders) are given, create a vector of zeros
    m, n = A.shape
    if p is None:
        p = Matrix(m, 1, lambda i, j : 0)

    # Verify the dimensions
    A_rows, A_cols = A.shape
    B_rows, B_cols = B.shape

    if (A_cols != B_cols) or (p.rows != A.rows):
        return ValueError('Dimension mismatch.')

    # Step 1: Find the inverse of A by solving A * x = e mod p
    # for each canonical generator e of the target space of A.
    # In other words, use the identity matrix to solve A * A_inv = I
    # by solving column-for-column.
    # Note: Only a one-sided inverse exists, A_inv * A = I is invalid.
    A_inv = []
    I_m = Matrix.eye(m)
    for i in range(m):
        identity_col = I_m[:, i]
        inv_col = solve(A, identity_col, p)

        # If a solution is found, append it
        if inv_col is not None:
            A_inv.append(inv_col)

    # Create the inverse of A from the list of column matrices
    A_inv = Matrix([[entry for entry in col] for col in A_inv]).T

    # Step 2: Multiply the inverse of A and B to form the solution
    # to the problem, the matrix X.
    X = B * A_inv

    return X


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose = True)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
from abelian import LCA, LCAFunc, HomLCA, voronoi


def is_close(list1, list2):
    return sum(abs(i-j) for (i,j) in zip(list1, list2)) < 10e-10

class TestLCAFunc:

    def test_natural_transversal(self):
        """
        Test transversal from T^2 to R^2.
        """

        T = LCA([1], [False])
        R = LCA([0], [False])
        A = [[0, 1], [1, -0.5]]
        epimorphism = HomLCA(A, source=R ** 2, target=T ** 2)
        sigma = voronoi(epimorphism, 2)

        assert is_close(sigma([0,0]), [0,0])
        assert is_close(sigma([0.8, 0.8]), [-0.3, -0.2])
        assert is_close(sigma([0.4, 0.8]), [0, 0.4])
        assert is_close(sigma([0.8, 0.4]), [0.3, -0.2])


if __name__ == '__main__':
    tests = TestLCAFunc()
    tests.test_natural_transversal()


#!/usr/bin/env python
# -*- coding: utf-8 -*-


import inspect
import importlib
import os
import sys

here = os.path.dirname(os.path.realpath(__file__))

def build_autosummaries(module_name, class_overview = True, functions = True, classes = True):
    """

    Parameters
    ----------
    module_name
        Name of the module.
    class_overview
        If true, builds a table with every class.
    functions
        If true, builds a table with every function.
    classes
        If true, builds a table for every single class.

    Returns
        string
            A string with .rst.
    -------

    """

    out = []
    sys.path.insert(0, '/home/tommy/Desktop/abelian')
    module = importlib.import_module(module_name)
    print(module)

    if class_overview:
        classes_in_module = inspect.getmembers(module, inspect.isclass)

        out.append('Public classes')
        out.append('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        out.append('\n.. autosummary::\n')

        for name, cls in classes_in_module:
            out.append('    ~' + cls.__module__ + '.' + name)
        out.append('')

    if functions:
        functions_in_module = inspect.getmembers(module, inspect.isfunction)

        out.append('Public functions')
        out.append('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        out.append('\n.. autosummary::\n')
        for name, function in functions_in_module:
            out.append('    ~' + function.__module__ + '.' + name)
        out.append('')

    if classes:
        classes_in_module = inspect.getmembers(module, inspect.isclass)

        out.append('Public classes (detailed)')
        out.append('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        for cls_name, cls in sorted(classes_in_module, reverse = True):

            # Find inheritance information
            bases = [basecls for basecls in cls.__bases__ if basecls != object]
            inherits = ', '.join([':class:`~{}`'.format(
                basecls.__module__ + '.' + basecls.__name__) for basecls in
                bases])

            header_clsname = cls.__module__ + '.' + cls_name
            header = ':class:`~{}`'.format(header_clsname)
            # Add inheritance if it exists
            out.append(header)
            out.append('^'*(25 +len(header)))
            if len(inherits) > 0:
                out.append(' (inherits from: {} )'.format(inherits))
            out.append('\n.. autosummary::\n')
            members = inspect.getmembers(cls)

            out.append('    ~' + header_clsname)
            for member_name, member in members:

                # If it's a function (normal method) or @classmethod
                if (inspect.isfunction(member) or
                        (inspect.ismethod(member) and member.__self__ is cls)):
                    print(member_name, member, type(member))

                    # If inherited, continue
                    if member_name not in list(cls.__dict__.keys()):
                        continue

                    if member_name[0] == '_' and member_name[1] != '_':
                        # Private method, continue
                        continue

                    out.append('    ~' + cls.__module__ + '.' + cls_name + '.' + member_name)

            out.append('  ')
        out.append('')

    return '\n'.join(out)


if True:# __name__ == '__main__':
    out = build_autosummaries('abelian')
    print(out)

    filename = 'autodoc_overview.rst'
    with open(os.path.join(here, filename), 'w', encoding = 'utf-8') as file:
        file.write(out)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
from random import randint as ri
from abelian import LCA, HomLCA
from sympy import Matrix

def random_from_list(number, list_to_take_from):
    """
    Draw several random values from the same list.
    """
    return [random.choice(list_to_take_from) for i in range(number)]

class TestHomLCA:

    @classmethod
    def setup_class(cls):
        """
        Setup a random phi:H -> G:
        """
        m, n = ri(2, 5), ri(2, 5)
        cls.H = LCA(random_from_list(n, [0,0,0,0,0,0]))
        cls.G = LCA(random_from_list(m, [0, 0, 0, 5, 8, 9, 10]))
        A = Matrix(m, n, lambda i,j : ri(-5,5))
        cls.phi = HomLCA(A, source = cls.H, target = cls.G)

    def test_homomorphism_property(self):
        """
        Test the homomorphism property.
        """
        H = self.H
        G = self.G
        phi = self.phi

        x = Matrix([ri(-9, 9) for i in range(len(H))])
        y = Matrix([ri(-9, 9) for i in range(len(H))])

        assert phi(x + y) == G(phi(x) + phi(y))

    def test_morphism_from_trivial(self):
        """
        Test a homomorphism from Z_1.
        """
        k = ri(1, 3)
        trivial_source = LCA(random_from_list(k, [1, 1, 1, 1]))
        G = self.G

        A = Matrix(len(G), k, lambda i,j : ri(-5, 5))
        phi_triv = HomLCA(A, source = trivial_source, target = G)

    def test_identity_morphism(self):
        """
        Test identity property of a morphism.
        """
        H = self.H
        G = self.G
        phi = self.phi

        Id_G = HomLCA.identity(G)
        Id_H = HomLCA.identity(H)

        assert (phi * Id_H).equal(phi)
        assert (Id_G * phi).equal(phi)

    def test_zero_morphism(self):
        """
        Test identity property of a morphism.
        """
        H = self.H
        G = self.G
        zero = HomLCA.zero(source = H, target = G)

        H_len = len(H)
        row_vector = Matrix(1, len(H), lambda i, j: 0)
        first = HomLCA(row_vector, source = H, target = LCA.trivial())

        col_vector = Matrix(len(G), 1, lambda i, j: 0)
        second = HomLCA(col_vector, source=LCA.trivial(), target=G)

        assert (second * first) == zero



if __name__ == '__main__':
    t = TestHomLCA()
    t.setup_class()
    t.test_homomorphism_property()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pkg_resources


__version__ = '0.2.0'
__author__ = 'Tommy Odland'

from abelian.linalg.solvers import solve
from abelian.linalg.factorizations import smith_normal_form, hermite_normal_form
from abelian.groups import LCA
from abelian.morphisms import HomLCA
from abelian.functions import LCAFunc, voronoi

__all__ = ['LCA', 'HomLCA', 'LCAFunc', 'voronoi']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
from sympy import Matrix
from random import randint as ri
from abelian.morphisms import HomLCA

def random_zero_heavy(low, high):
    """
    Draw a random number, with approx 50% probability of zero.
    """
    return random.choice(list(range(low, high)) + [0]*(high - low))

class TestHomLCA:

    @classmethod
    def setup_class(cls):
        """Setup two random homomorphisms phi and psi.
        """
        cls.m, cls.n = 3, 3
        cls.phi = HomLCA(Matrix(cls.m, cls.n, lambda i, j: ri(-5, 5)))
        cls.psi = HomLCA(Matrix(cls.m, cls.n, lambda i, j: ri(-5, 5)))

    def test_stack_horizonally(self):
        """
        Test the horizontal stacking property.
        """
        # Create random group elements (inputs)
        x = Matrix([ri(-5, 5) for i in range(self.n)])
        y = Matrix([ri(-5, 5) for i in range(self.n)])

        x_over_y = x.col_join(y)
        stacked = self.phi.stack_horiz(self.psi)
        assert stacked(x_over_y) == self.phi(x) + self.psi(y)

    def test_stack_vertically(self):
        """
        Test the vertical stacking property.
        """

        # Create random group elements (inputs)
        x = Matrix([ri(-5, 5) for i in range(self.n)])

        stacked = self.phi.stack_vert(self.psi)
        assert stacked(x) == self.phi(x).col_join(self.psi(x))

    def test_stack_diagonally(self):
        """
        Test the diagonal stacking property.
        """

        # Create random group elements (inputs)
        x = Matrix([ri(-5, 5) for i in range(self.n)])
        y = Matrix([ri(-5, 5) for i in range(self.n)])

        x_over_y = x.col_join(y)
        stacked = self.phi.stack_diag(self.psi)

        assert stacked(x_over_y) == self.phi(x).col_join(self.psi(y))

    def test_call_order(self):
        """
        Test that (a \circ b)(x) == a (b(x)).
        """
        # Create random group elements (inputs)
        x = Matrix([ri(-5, 5) for i in range(self.n)])
        composed = (self.phi * self.psi)
        assert composed(x) == self.phi(self.psi(x))


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
from sympy import Matrix
from random import randint as ri
from abelian.morphisms import HomLCA

def random_zero_heavy(low, high):
    """
    Draw a random number, with approx 50% probability of zero.
    """
    return random.choice(list(range(low, high)) + [0]*(high - low))

class TestSNF:

    @classmethod
    def setup_class(cls):
        """
        Setup a homomorphism.
        """
        m, n = ri(2, 4), ri(2, 4)
        A = Matrix(m, n, lambda i, j: random_zero_heavy(-9, 9))
        target = Matrix(m, 1, lambda i, j: random_zero_heavy(5, 50))

        cls.target_vector = target
        cls.phi = HomLCA(A, target = target)


    def test_source_projection_two_ways(self):
        """
        Test that phi.project to source == phi.coimage.project_to_source
        """
        # Project to source
        phi = self.phi.project_to_source()
        coimage = self.phi.coimage().project_to_source()
        assert phi.source == coimage.source


    def test_project_to_source(self):
        """
        Test the orders.
        """

        # Project to source
        phi = self.phi.project_to_source()

        # Retrieve the orders as a vector
        orders = Matrix(phi.source.orders)

        # Assert that the columns times the orders are zero
        assert phi.evaluate(orders) == Matrix(phi.target.orders) * 0


    def test_image_coimage_factorization(self):
        """
        Test the image/coimage factorization.
        """

        # Compute the image morphism and the coimage morphism
        image = self.phi.image().remove_trivial_groups()
        coimage = self.phi.coimage().remove_trivial_groups()

        # Asser that the coimage/image factorization holds
        factorization = (image * coimage).project_to_target()
        original = (self.phi).project_to_target()
        assert factorization == original


    def test_kernel(self):
        """
        The that the kernel and the morphism is zero.
        """
        # Compute the kernel
        kernel = self.phi.kernel()
        phi_ker = (self.phi * kernel).project_to_target()
        zero_morphism = HomLCA.zero(target = self.phi.target,
                                    source = kernel.source)
        assert phi_ker == zero_morphism


    def test_cokernel(self):
        """
        Test that the composition of the cokernel and the morphism is zero.
        """
        cokernel = self.phi.cokernel()
        coker_phi = (cokernel * self.phi).project_to_target()
        zero_morphism = HomLCA.zero(source = self.phi.source,
                                    target = cokernel.target)
        assert coker_phi == zero_morphism


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module tests free-to-free kernels, cokernels images and coimages.
It also test generation of elements by increasing max norm.
All tese are stochastics, of dimensions/rank 3-5 typically.
"""

import itertools
from random import randint as ri
from random import choice
from sympy import Matrix

from abelian.linalg.free_to_free import free_coimage, free_cokernel, \
    free_image, free_kernel, free_quotient, elements_of_maxnorm_FGA, \
    elements_of_maxnorm
from abelian.linalg.utils import vector_mod_vector
from abelian.linalg.free_to_free import mod

class TestElementsGeneratorFree:
    """
    Test the free elements generator,
    which generates elements of Z^r with increasing max-norm.
    """

    def test_elements_of_maxnorm_num_elements(self):
        """
        Verify the number of elements.
        """

        # Random parameter values
        dim = ri(1, 5)
        normvalue = ri(3, 5)

        # Theoretical value
        theoretical_value = (2*normvalue + 1)**dim - (2*normvalue - 1)**dim

        # Actual value
        generated = list(elements_of_maxnorm(dim, normvalue))
        assert len(set(generated)) == theoretical_value


    def test_elements_of_maxnorm_increasing_norm(self):
        """
        Verify that the norm is increasing.
        """

        # Random parameter values
        dim = ri(1, 4)
        normvalue = ri(2, 4)

        # Generate elements up to a norm
        generated = []
        for k in range(normvalue):
            generated += list(elements_of_maxnorm(dim, k))

        # Define the maximum norm
        norm = lambda v : max(abs(k) for k in v)

        # Assert that the norm is increasing
        assert all(norm(a)<= norm(b) for a, b in
                   zip(generated[:-1], generated[1:]))


def naive_FGA_elements_by_norm(orders, maxnorm_value):
    """
    Naively generated every element in Z_p such
    that maxnorm(element) = maxnorm_value.

    This algorithm goes through a cartesian product, generates
    everything, projects and checks the norm.
    """

    # Wrap around max norm
    def maxnorm_in_finite(a, b):
        return tuple([min(abs(i), abs(i - j)) for i, j in zip(a, b)])

    # Argument into cartesian product and storage for yielded values
    prodarg = [range(-maxnorm_value, maxnorm_value + 1) for o in orders]
    yielded = set()

    # Take Cartesian products
    for prod in itertools.product(*prodarg):
        # Project argument
        prod = mod(prod, tuple(orders))
        # If not yielded and correct norm
        if (prod not in yielded) and \
                (max(maxnorm_in_finite(prod, tuple(orders))) == maxnorm_value):
            yielded.add(prod)
            yield prod

class TestElementsGeneratorFGA:

    def test_free_equality(self):
        """
        Verify the number of elements.
        """

        # Random parameter values
        dim = ri(1, 4)
        normvalue = ri(2, 4)

        # Generate with free
        ret1 = list(elements_of_maxnorm(dim, normvalue))
        ret2 = list(elements_of_maxnorm_FGA(dim*[0], normvalue))

        ret1.sort()
        ret2.sort()

        assert ret1 == ret2

    def test_VS_naive_generator(self):
        """
        Test vs the naive generator.
        Make sure the returned values are equal.
        """
        normvalue = ri(0, 4)
        choices = [0]*5 + [4,5,12,13]
        orders = [choice(choices) for k in range(ri(1,4))]

        # Return from naive
        ret1 = list(naive_FGA_elements_by_norm(orders, normvalue))

        # Return from algorithm to test against
        ret2 = list(elements_of_maxnorm_FGA(orders, normvalue))

        # Sort and compare
        ret1.sort()
        ret2.sort()
        assert ret1 == ret2






class TestFreeToFree:
    @classmethod
    def setup_class(cls):
        m, n = ri(3, 5), ri(3, 5)
        A = Matrix(m, n, lambda i, j: ri(-9, 9))
        cls.A = A

        # Compute ker, coker, im, coim and quotient group
        cls.A_ker = free_kernel(A)
        cls.A_coker = free_cokernel(A)
        cls.A_im = free_image(A)
        cls.A_coim = free_coimage(A)
        cls.A_quotient = free_quotient(A)

    def test_im_coim(self):
        """
        Test the image/coimage factorization of A.
        """
        assert (self.A_im * self.A_coim == self.A)

    def test_kernel(self):
        """
        Test that the composition of A and ker(A) is zero.
        """
        sum_entries = sum(abs(entry) for entry in self.A * self.A_ker)
        assert (sum_entries == 0)

    def test_cokernel(self):
        """
        Test that the composition of A and ker(A) is zero in the target group.
        """
        prod = self.A_coker * self.A
        m, n = prod.shape
        ones = Matrix(n, 1, list(1 for i in range(n)))
        prod_in_target_grp = vector_mod_vector(prod * ones, self.A_quotient)
        assert (prod_in_target_grp == self.A_quotient * 0)
